import copy
from logging import getLogger
from os.path import abspath, dirname, join, pardir

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from loren_frank_data_processing import (Animal, get_all_multiunit_indicators,
                                         get_all_spike_indicators,
                                         get_interpolated_position_dataframe,
                                         get_LFPs, get_spikes_dataframe,
                                         get_trial_time, make_neuron_dataframe,
                                         make_tetrode_dataframe)
from loren_frank_data_processing.position import make_track_graph
from ripple_detection import (Kay_ripple_detector, filter_ripple_band,
                              get_multiunit_population_firing_rate,
                              multiunit_HSE_detector)
from spectral_connectivity import Connectivity, Multitaper
from statsmodels.tsa.tsatools import lagmat

logger = getLogger(__name__)


SAMPLING_FREQUENCY = 500
_MARKS = ['channel_1_max', 'channel_2_max', 'channel_3_max', 'channel_4_max']
BRAIN_AREAS = ['CA1', 'CA2', 'CA3']

# Data directories and definitions
ROOT_DIR = join(abspath(dirname(__file__)), pardir)
RAW_DATA_DIR = join(ROOT_DIR, 'Raw-Data')

ANIMALS = {
    'bon': Animal(directory=join(RAW_DATA_DIR, 'Bond'), short_name='bon'),
}

logger = getLogger(__name__)

_MARKS = ['channel_1_max', 'channel_2_max', 'channel_3_max', 'channel_4_max']


def get_labels(times, time):
    ripple_labels = pd.DataFrame(np.zeros_like(time, dtype=np.int), index=time,
                                 columns=['replay_number'])
    for replay_number, start_time, end_time in times.itertuples():
        ripple_labels.loc[start_time:end_time] = replay_number

    return ripple_labels


def estimate_ripple_band_power(lfps, sampling_frequency):
    m = Multitaper(lfps.values, sampling_frequency=sampling_frequency,
                   time_halfbandwidth_product=1,
                   time_window_duration=0.020,
                   time_window_step=0.020,
                   start_time=lfps.index[0].total_seconds())
    c = Connectivity.from_multitaper(m)
    closest_200Hz_freq_ind = np.argmin(np.abs(c.frequencies - 200))
    power = c.power()[..., closest_200Hz_freq_ind, :].squeeze() + np.spacing(1)
    n_samples = int(0.020 * sampling_frequency)
    index = lfps.index[np.arange(1, power.shape[0] * n_samples + 1, n_samples)]
    power = pd.DataFrame(power, index=index)
    return power.reindex(lfps.index)


def get_adhoc_ripple(epoch_key, tetrode_info, position_time):
    LFP_SAMPLING_FREQUENCY = 1500
    position_info = (
        get_interpolated_position_dataframe(epoch_key, ANIMALS)
        .dropna(subset=['linear_position', 'speed']))
    speed = position_info['speed']
    time = position_info.index

    if ~np.all(np.isnan(tetrode_info.validripple.astype(float))):
        tetrode_keys = tetrode_info.loc[
            (tetrode_info.validripple == 1)].index
    else:
        is_brain_areas = (
            tetrode_info.area.astype(str).str.upper().isin(BRAIN_AREAS))
        tetrode_keys = tetrode_info.loc[is_brain_areas].index

    ripple_lfps = get_LFPs(tetrode_keys, ANIMALS).reindex(time)
    ripple_filtered_lfps = pd.DataFrame(
        np.stack([filter_ripple_band(
            ripple_lfps.values[:, ind], sampling_frequency=1500)
            for ind in np.arange(ripple_lfps.shape[1])], axis=1),
        index=ripple_lfps.index)
    ripple_times = Kay_ripple_detector(
        time, ripple_lfps.values, speed.values, LFP_SAMPLING_FREQUENCY,
        zscore_threshold=2.0, close_ripple_threshold=np.timedelta64(0, 'ms'),
        minimum_duration=np.timedelta64(15, 'ms'))

    ripple_times.index = ripple_times.index.rename('replay_number')
    ripple_labels = get_labels(ripple_times, position_time)
    is_ripple = ripple_labels > 0
    ripple_times = ripple_times.assign(
        duration=lambda df: (df.end_time - df.start_time).dt.total_seconds())

    ripple_power = estimate_ripple_band_power(
        ripple_lfps, LFP_SAMPLING_FREQUENCY)
    interpolated_ripple_power = ripple_power.interpolate()

    ripple_power_change = interpolated_ripple_power.transform(
        lambda df: df / df.mean())
    ripple_power_zscore = np.log(interpolated_ripple_power).transform(
        lambda df: (df - df.mean()) / df.std())

    return dict(ripple_times=ripple_times,
                ripple_labels=ripple_labels,
                ripple_filtered_lfps=ripple_filtered_lfps,
                ripple_power=ripple_power,
                ripple_lfps=ripple_lfps,
                ripple_power_change=ripple_power_change,
                ripple_power_zscore=ripple_power_zscore,
                is_ripple=is_ripple)


def get_adhoc_multiunit(speed, tetrode_info, time_function):
    time = time_function()
    is_brain_areas = (
        tetrode_info.area.astype(str).str.upper().isin(BRAIN_AREAS))
    tetrode_info = tetrode_info.loc[is_brain_areas]

    multiunit = (get_all_multiunit_indicators(
        tetrode_info.index, ANIMALS, time_function)
        .sel(features=_MARKS)
        .reindex({'time': time}))
    multiunit_spikes = (np.any(~np.isnan(multiunit.values), axis=1)
                        ).astype(np.float)
    multiunit_firing_rate = pd.DataFrame(
        get_multiunit_population_firing_rate(
            multiunit_spikes, SAMPLING_FREQUENCY), index=time,
        columns=['firing_rate'])
    multiunit_rate_change = multiunit_firing_rate.transform(
        lambda df: df / df.mean())
    multiunit_rate_zscore = np.log(multiunit_firing_rate).transform(
        lambda df: (df - df.mean()) / df.std())

    multiunit_high_synchrony_times = multiunit_HSE_detector(
        time, multiunit_spikes, speed.values, SAMPLING_FREQUENCY,
        minimum_duration=np.timedelta64(15, 'ms'), zscore_threshold=2.0,
        close_event_threshold=np.timedelta64(0, 'ms'))
    multiunit_high_synchrony_times.index = (
        multiunit_high_synchrony_times.index.rename('replay_number'))
    multiunit_high_synchrony_labels = get_labels(
        multiunit_high_synchrony_times, time)
    is_multiunit_high_synchrony = multiunit_high_synchrony_labels > 0
    multiunit_high_synchrony_times = multiunit_high_synchrony_times.assign(
        duration=lambda df: (df.end_time - df.start_time).dt.total_seconds())

    return dict(
        multiunit=multiunit,
        multiunit_spikes=multiunit_spikes,
        multiunit_firing_rate=multiunit_firing_rate,
        multiunit_high_synchrony_times=multiunit_high_synchrony_times,
        multiunit_high_synchrony_labels=multiunit_high_synchrony_labels,
        multiunit_rate_change=multiunit_rate_change,
        multiunit_rate_zscore=multiunit_rate_zscore,
        is_multiunit_high_synchrony=is_multiunit_high_synchrony)


def get_spikes(neuron_info, time_function):
    time = time_function()
    neuron_info = neuron_info.loc[
        (neuron_info.numspikes > 100) &
        neuron_info.area.isin(BRAIN_AREAS) &
        (neuron_info.type == 'principal')]
    spikes = get_all_spike_indicators(
        neuron_info.index, ANIMALS, time_function).reindex(time)
    spike_times = [get_spikes_dataframe(neuron_key, ANIMALS)
                   for neuron_key in neuron_info.index]

    return spikes, spike_times


def load_data(epoch_key):
    logger.info('Loading data...')
    time = get_trial_time(epoch_key, ANIMALS)
    time = (pd.Series(np.ones_like(time, dtype=np.float), index=time)
            .resample('2ms').mean()
            .index)

    def _time_function(*args, **kwargs):
        return time

    position_info = (
        get_interpolated_position_dataframe(
            epoch_key, ANIMALS, _time_function)
        .dropna(subset=['linear_position', 'speed']))

    time = position_info.index
    speed = position_info['speed']

    neuron_info = make_neuron_dataframe(ANIMALS).xs(
        epoch_key, drop_level=False)
    spikes, spike_times = get_spikes(neuron_info, _time_function)
    tetrode_info = make_tetrode_dataframe(ANIMALS, epoch_key=epoch_key)

    track_graph, _ = make_track_graph(epoch_key, ANIMALS)

    logger.info('Finding multiunit high synchrony events...')
    adhoc_multiunit = get_adhoc_multiunit(
        speed, tetrode_info, _time_function)

    logger.info('Finding ripple times...')
    adhoc_ripple = get_adhoc_ripple(epoch_key, tetrode_info, time)
    return {
        'position_info': position_info,
        'tetrode_info': tetrode_info,
        'neuron_info': neuron_info,
        'spikes': spikes,
        'spike_times': spike_times,
        'track_graph': track_graph,
        'sampling_frequency': SAMPLING_FREQUENCY,
        **adhoc_ripple,
        **adhoc_multiunit,
    }


def plot_detector_debug(time_ind, data, replay_detector, detector_results,
                        figsize=(10, 7.5)):
    fig, axes = plt.subplots(7, 1,
                             figsize=figsize,
                             sharex=True,
                             constrained_layout=True,
                             gridspec_kw={
                                 'height_ratios': [3, 1, 3, 3, 3, 1, 1]})

    place_fields = (
        replay_detector._spiking_likelihood
        .keywords['place_conditional_intensity']).squeeze()
    time = data['spikes'].iloc[time_ind].index / np.timedelta64(1, 's')

    place_field_max = np.argmax(place_fields, axis=0)
    linear_position_order = place_field_max.argsort(axis=0).squeeze()
    spike_time_ind, neuron_ind = np.nonzero(
        np.asarray(data['spikes'].iloc[time_ind])[:, linear_position_order])

    axes[0].scatter(time[spike_time_ind], neuron_ind,
                    clip_on=False, s=10, color='black', marker='|')
    axes[0].set_ylim((0, place_fields.shape[1]))
    axes[0].set_yticks((0, place_fields.shape[1]))
    axes[0].set_ylabel('Cells')
    axes[0].fill_between(
        time, np.ones_like(time) * place_fields.shape[1],
        where=detector_results.isel(time=time_ind).replay_probability >= 0.8,
        color='lightgrey', zorder=-1, alpha=0.6, step='pre')
    axes[0].fill_between(
        time, np.ones_like(time) * place_fields.shape[1],
        where=data['ripple_labels'].iloc[time_ind].values.squeeze() > 0,
        color='red', zorder=-2, alpha=0.1, step='pre')

    detector_results.causal_posterior.isel(
        time=time_ind).sum("position").sel(state="Non-Local").plot(
        x='time', ax=axes[1], color='black', clip_on=False, linewidth=0.5,
        alpha=0.75
    )
    detector_results.isel(time=time_ind).replay_probability.plot(
        x='time', ax=axes[1], color='black', clip_on=False)
    axes[1].set_ylabel('Prob.')
    axes[1].set_xlabel("")
    axes[1].set_ylim((0, 1))

    cmap = copy.copy(plt.get_cmap('bone_r'))
    cmap.set_bad(color="lightgrey", alpha=1.0)
    (detector_results
     .isel(time=time_ind)
     .causal_posterior
     .sum('state')
     .where(replay_detector.is_track_interior_)
     .plot(x='time', y='position', ax=axes[2], cmap=cmap,
           vmin=0.0, vmax=0.02))
    axes[2].scatter(data['position_info'].iloc[time_ind].index /
                    np.timedelta64(1, 's'),
                    data['position_info'].iloc[time_ind].linear_position,
                    s=1,
                    color='magenta', clip_on=False)
    axes[2].set_xlabel("")
    axes[2].set_ylabel('Position')

    (detector_results
     .isel(time=time_ind)
     .acausal_posterior
     .sum('state')
     .where(replay_detector.is_track_interior_)
     .plot(x='time', y='position', ax=axes[3], cmap=cmap,
           vmin=0.0, vmax=0.02))
    axes[3].scatter(data['position_info'].iloc[time_ind].index /
                    np.timedelta64(1, 's'),
                    data['position_info'].iloc[time_ind].linear_position,
                    s=1,
                    color='magenta', clip_on=False)
    axes[3].set_xlabel("")
    axes[3].set_ylabel('Position')

    (detector_results
     .isel(time=time_ind, state=1)
     .likelihood
     .where(replay_detector.is_track_interior_)
     .plot(x='time', y='position', ax=axes[4], vmin=0.75, vmax=1.0,
           cmap='viridis'))
    axes[4].scatter(data['position_info'].iloc[time_ind].index /
                    np.timedelta64(1, 's'),
                    data['position_info'].iloc[time_ind].linear_position,
                    s=1,
                    color='magenta', clip_on=False)
    axes[4].set_xlabel("")
    axes[4].set_title("")
    axes[4].set_ylabel('Position')

    speed = np.asarray(data['position_info'].iloc[time_ind].speed).squeeze()
    axes[5].fill_between(time, speed, color='grey')
    axes[5].axhline(4, color='black', linestyle='--', linewidth=1)
    axes[5].set_ylabel('Speed')

    lagged_speed = np.asarray(
        lagmat(data['position_info'].speed, maxlag=1, use_pandas=True)
        .iloc[time_ind]).squeeze()
    probablity_replay = replay_detector.replay_state_transition_(lagged_speed)
    axes[6].plot(time, probablity_replay[:, 0])
    axes[6].twinx().plot(time, probablity_replay[:, 1], color="orange")

    sns.despine(offset=5)


def plot_detector(time_ind, data, replay_detector, detector_results,
                  figsize=(10.5, 6.0)):
    fig, axes = plt.subplots(4, 1,
                             figsize=figsize,
                             sharex=True,
                             constrained_layout=True,
                             gridspec_kw={
                                 'height_ratios': [3, 1, 3, 1]})

    place_fields = (
        replay_detector._spiking_likelihood
        .keywords['place_conditional_intensity']).squeeze()
    time = data['spikes'].iloc[time_ind].index / np.timedelta64(1, 's')

    place_field_max = np.argmax(place_fields, axis=0)
    linear_position_order = place_field_max.argsort(axis=0).squeeze()
    spike_time_ind, neuron_ind = np.nonzero(
        np.asarray(data['spikes'].iloc[time_ind])[:, linear_position_order])

    axes[0].scatter(time[spike_time_ind], neuron_ind,
                    clip_on=False, s=10, color='black', marker='|')
    axes[0].set_ylim((0, place_fields.shape[1]))
    axes[0].set_yticks((0, place_fields.shape[1]))
    axes[0].set_ylabel('Cells')
    axes[0].fill_between(
        time, np.ones_like(time) * place_fields.shape[1],
        where=detector_results.isel(time=time_ind).replay_probability >= 0.8,
        color='lightgrey', zorder=-1, alpha=0.6, step='pre')
    axes[0].fill_between(
        time, np.ones_like(time) * place_fields.shape[1],
        where=data['ripple_labels'].iloc[time_ind].values.squeeze() > 0,
        color='red', zorder=-2, alpha=0.1, step='pre')

    detector_results.isel(time=time_ind).replay_probability.plot(
        x='time', ax=axes[1], color='black', clip_on=False)
    axes[1].set_ylabel('Prob.')
    axes[1].set_xlabel("")
    axes[1].set_ylim((0, 1))

    cmap = copy.copy(plt.get_cmap('bone_r'))
    cmap.set_bad(color="lightgrey", alpha=1.0)

    (detector_results
     .isel(time=time_ind)
     .acausal_posterior
     .sum('state')
     .where(replay_detector.is_track_interior_)
     .plot(x='time', y='position', ax=axes[2], cmap=cmap,
           vmin=0.0, vmax=0.02))
    axes[2].scatter(data['position_info'].iloc[time_ind].index /
                    np.timedelta64(1, 's'),
                    data['position_info'].iloc[time_ind].linear_position,
                    s=1,
                    color='magenta', clip_on=False)
    axes[2].set_xlabel("")
    axes[2].set_ylabel('Position')

    speed = np.asarray(data['position_info'].iloc[time_ind].speed).squeeze()
    axes[3].fill_between(time, speed, color='grey')
    axes[3].axhline(4, color='black', linestyle='--', linewidth=1)
    axes[3].set_ylabel('Speed')

    sns.despine(offset=5)



from logging import getLogger

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
import sklearn
import xarray as xr
from replay_identification.bins import (atleast_2d, get_centers, get_grid,
                                        get_observed_position_bin,
                                        get_track_grid, get_track_interior)
from replay_identification.core import (_acausal_classifier,
                                        _acausal_classifier_gpu,
                                        _causal_classifier,
                                        _causal_classifier_gpu,
                                        check_converged)
from replay_identification.discrete_state_transition import (
    estimate_discrete_state_transition,
    infer_discrete_state_transition_from_training_data,
    make_discrete_state_transition_from_diagonal)
from replay_identification.movement_state_transition import (
    empirical_movement, random_walk, random_walk_on_track_graph)
from replay_identification.multiunit_likelihood import (
    NumbaKDE, fit_multiunit_likelihood)
from replay_identification.multiunit_likelihood_integer import \
    fit_multiunit_likelihood_integer
from replay_identification.multiunit_likelihood_integer_gpu import \
    fit_multiunit_likelihood_gpu
from replay_identification.spiking_likelihood import fit_spiking_likelihood
from sklearn.base import BaseEstimator

logger = getLogger(__name__)
sklearn.set_config(print_changed_only=False)

_DISCRETE_DIAGONAL = np.asarray([0.999, 0.98])
_DEFAULT_CLUSTERLESS_MODEL_KWARGS = {
    'density_model': NumbaKDE,
    'model_kwargs': dict(
        bandwidth=np.array([24.0, 24.0, 24.0, 24.0, 6.0, 6.0])),
    'occupancy_marginal_model': NumbaKDE,
    'occupancy_kwargs': dict(bandwidth=np.array([6.0, 6.0]))
}


class _BaseDetector(BaseEstimator):
    def __init__(
        self,
        place_bin_size=2.0,
        position_range=None,
        track_graph=None,
        edge_order=None,
        edge_spacing=None,
        continuous_state_transition_type='random_walk',
        random_walk_variance=6.0,
        discrete_state_transition_type='infer_from_training_data',
        discrete_transition_diagonal=_DISCRETE_DIAGONAL,
        is_track_interior=None,
        infer_track_interior=True,
    ):
        self.place_bin_size = place_bin_size
        self.position_range = position_range
        self.track_graph = track_graph
        self.edge_order = edge_order
        self.edge_spacing = edge_spacing
        self.continuous_state_transition_type = continuous_state_transition_type
        self.random_walk_variance = random_walk_variance
        self.discrete_state_transition_type = discrete_state_transition_type
        self.discrete_transition_diagonal = discrete_transition_diagonal
        self.is_track_interior = is_track_interior
        self.infer_track_interior = infer_track_interior

    def fit_place_grid(
            self,
            position=None,
            track_graph=None,
            edge_order=None,
            edge_spacing=None,
            infer_track_interior=True
    ):
        position = atleast_2d(np.asarray(position))
        self.track_graph = track_graph
        if self.track_graph is None:
            (self.edges_,
             self.place_bin_edges_,
             self.place_bin_centers_,
             self.centers_shape_
             ) = get_grid(position, self.place_bin_size, self.position_range,
                          self.infer_track_interior)

            self.infer_track_interior = infer_track_interior

            if self.is_track_interior is None and self.infer_track_interior:
                self.is_track_interior_ = get_track_interior(
                    position, self.edges_)
            elif (self.is_track_interior is None and
                  not self.infer_track_interior):
                self.is_track_interior_ = np.ones(
                    self.centers_shape_, dtype=np.bool)
        else:
            (
                self.place_bin_centers_,
                self.place_bin_edges_,
                self.is_track_interior_,
                self.distance_between_nodes_,
                self.centers_shape_,
                self.edges_,
                self.track_graph_with_bin_centers_edges_,
                self.original_nodes_df_,
                self.place_bin_edges_nodes_df_,
                self.place_bin_centers_nodes_df_,
                self.nodes_df_
            ) = get_track_grid(self.track_graph, edge_order,
                               edge_spacing, self.place_bin_size)

        return self

    def fit_discrete_state_transition(
            self,
            discrete_transition_diagonal=None,
            is_training=None):
        logger.info('Fitting discrete state transition...')
        if discrete_transition_diagonal is not None:
            self.discrete_transition_diagonal = discrete_transition_diagonal

        if self.discrete_state_transition_type == 'infer_from_training_data':
            self.discrete_state_transition_ = infer_discrete_state_transition_from_training_data(
                ~is_training)
        elif self.discrete_state_transition_type == 'make_from_user_specified_diagonal':
            self.discrete_state_transition_ = make_discrete_state_transition_from_diagonal(
                discrete_transition_diagonal)
        else:
            raise NotImplementedError

    def plot_discrete_state_transition(
            self, state_names=None, cmap='Oranges', ax=None,
            convert_to_seconds=False, sampling_frequency=1):

        if ax is None:
            ax = plt.gca()

        if state_names is None:
            state_names = ['Local', 'Non-Local']

        if convert_to_seconds:
            discrete_state_transition = (
                1 / (1 - self.discrete_state_transition_)) / sampling_frequency
            vmin, vmax, fmt = 0.0, None, '0.03f'
            label = 'Seconds'
        else:
            discrete_state_transition = self.discrete_state_transition_
            vmin, vmax, fmt = 0.0, 1.0, '0.03f'
            label = 'Probability'

        sns.heatmap(data=discrete_state_transition,
                    vmin=vmin, vmax=vmax, annot=True, fmt=fmt, cmap=cmap,
                    xticklabels=state_names, yticklabels=state_names, ax=ax,
                    cbar_kws={'label': label})
        ax.set_ylabel('Previous State', fontsize=12)
        ax.set_xlabel('Current State', fontsize=12)
        ax.set_title('Discrete State Transition', fontsize=16)

    def fit_continuous_state_transition(
        self,
        position=None,
        is_training=None,
    ):
        logger.info('Fitting continuous state transition...')
        if self.continuous_state_transition_type == 'empirical':
            if is_training is None:
                n_time = position.shape[0]
                is_training = np.ones((n_time,), dtype=bool)
            if position is not None:
                position = atleast_2d(np.asarray(position))
            self.continuous_state_transition_ = empirical_movement(
                position,
                self.edges_,
                is_training,
                replay_speed=20)
        elif ((self.continuous_state_transition_type == 'random_walk') &
              (self.track_graph is not None)):
            place_bin_center_ind_to_node = np.asarray(
                self.place_bin_centers_nodes_df_.node_id)
            self.continuous_state_transition_ = random_walk_on_track_graph(
                self.place_bin_centers_,
                0.0,
                self.random_walk_variance,
                place_bin_center_ind_to_node,
                self.distance_between_nodes_
            )
        elif self.continuous_state_transition_type == 'random_walk':
            self.continuous_state_transition_ = random_walk(
                self.place_bin_centers_,
                self.random_walk_variance,
                self.is_track_interior_.ravel(order='F'),
                1)

    def plot_continuous_state_transition(self, ax=None):
        if ax is None:
            ax = plt.gca()
        ax.pcolormesh(self.continuous_state_transition_.T)

    def _get_results(
        self,
        position,
        likelihood,
        time=None,
        is_compute_acausal=True,
        use_gpu=False
    ):
        n_time = likelihood.shape[0]
        if time is None:
            time = np.arange(n_time)

        observed_position_bin = get_observed_position_bin(
            position,
            self.edges_,
            self.place_bin_centers_,
            self.is_track_interior_.ravel(order='F')
        )

        uniform = np.ones((self.place_bin_centers_.shape[0],))
        uniform[~self.is_track_interior_.ravel(order='F')] = 0.0
        uniform /= uniform.sum()

        likelihood[:, :, ~self.is_track_interior_.ravel(order='F')] = 0.0

        logger.info('Finding causal non-local probability and position...')
        if not use_gpu:
            (causal_posterior, state_probability,
             data_log_likelihood) = _causal_classifier(
                likelihood,
                self.continuous_state_transition_,
                np.repeat(self.discrete_state_transition_[
                          :, [1]].T, n_time, axis=0),
                observed_position_bin,
                uniform)
        else:
            (causal_posterior, state_probability,
             data_log_likelihood) = _causal_classifier_gpu(
                likelihood,
                self.continuous_state_transition_,
                np.repeat(self.discrete_state_transition_[
                          :, [1]].T, n_time, axis=0),
                observed_position_bin,
                uniform)

        n_position_dims = self.place_bin_centers_.shape[1]

        if (likelihood.shape[-1] > 1) & (n_position_dims > 1):
            likelihood_dims = ['time', 'state', 'x_position', 'y_position']
            posterior_dims = likelihood_dims
            coords = dict(
                time=time,
                x_position=get_centers(self.edges_[0]),
                y_position=get_centers(self.edges_[1]),
                state=['Local', 'Non-Local'],
            )
            likelihood_shape = (n_time, 2, *self.centers_shape_)
            posterior_shape = likelihood_shape
        elif (likelihood.shape[-1] > 1) & (n_position_dims == 1):
            likelihood_dims = ['time', 'state', 'position']
            posterior_dims = likelihood_dims
            coords = dict(
                time=time,
                position=get_centers(self.edges_[0]),
                state=['Local', 'Non-Local'],
            )
            likelihood_shape = (n_time, 2, *self.centers_shape_)
            posterior_shape = likelihood_shape
        else:
            likelihood_dims = ['time', 'state']
            posterior_dims = ['time', 'state', 'position']
            likelihood_shape = (n_time, 2)
            posterior_shape = (n_time, 2, *self.centers_shape_)
            coords = dict(
                time=time,
                position=get_centers(self.edges_[0]),
                state=['Local', 'Non-Local'],
            )

        try:
            results = xr.Dataset(
                {'causal_posterior': (
                    posterior_dims,
                    causal_posterior.reshape(posterior_shape).swapaxes(3, 2)),
                 'likelihood': (
                    likelihood_dims,
                    likelihood.reshape(likelihood_shape).swapaxes(3, 2))},
                coords=coords,
                attrs=dict(data_log_likelihood=data_log_likelihood))
        except np.AxisError:
            results = xr.Dataset(
                {'causal_posterior': (
                    posterior_dims,
                    causal_posterior.squeeze()),
                 'likelihood': (
                    likelihood_dims,
                    likelihood.squeeze())},
                coords=coords,
                attrs=dict(data_log_likelihood=data_log_likelihood))
        if is_compute_acausal:
            logger.info(
                'Finding acausal non-local probability and position...')

            if not use_gpu:
                acausal_posterior, state_probability = _acausal_classifier(
                    causal_posterior,
                    self.continuous_state_transition_,
                    np.repeat(self.discrete_state_transition_[
                              :, [1]].T, n_time, axis=0),
                    observed_position_bin,
                    uniform)
            else:
                acausal_posterior, state_probability = _acausal_classifier_gpu(
                    causal_posterior,
                    self.continuous_state_transition_,
                    np.repeat(self.discrete_state_transition_[
                              :, [1]].T, n_time, axis=0),
                    observed_position_bin,
                    uniform)

            try:
                results['acausal_posterior'] = (
                    posterior_dims,
                    acausal_posterior.reshape(posterior_shape).swapaxes(3, 2))
            except np.AxisError:
                results['acausal_posterior'] = (
                    posterior_dims,
                    acausal_posterior.squeeze())
        results['non_local_probability'] = (['time'], state_probability[:, 1])

        return results

    def estimate_parameters(self, fit_args, predict_args, tolerance=1E-4,
                            max_iter=10, estimate_state_transition=True,
                            estimate_likelihood=True):

        self.fit(**fit_args)
        results = self.predict(**predict_args)

        data_log_likelihoods = [results.data_log_likelihood]
        log_likelihood_change = np.inf
        converged = False
        increasing = True
        n_iter = 0

        logger.info(
            f'iteration {n_iter}, likelihood: {data_log_likelihoods[-1]}')

        while not converged and (n_iter < max_iter):
            if estimate_state_transition:
                self.discrete_state_transition_ = estimate_discrete_state_transition(
                    self, results)
            if estimate_likelihood:
                fit_args['is_training'] = 1 - results.non_local_probability
                fit_args['refit'] = True
                self.fit(**fit_args)

            results = self.predict(**predict_args)
            data_log_likelihoods.append(results.data_log_likelihood)
            log_likelihood_change = (
                data_log_likelihoods[-1] - data_log_likelihoods[-2])
            n_iter += 1

            converged, increasing = check_converged(
                data_log_likelihoods[-1], data_log_likelihoods[-2], tolerance)

            logger.info(
                f'iteration {n_iter}, '
                f'likelihood: {data_log_likelihoods[-1]}, '
                f'change: {log_likelihood_change}'
            )

        return results, data_log_likelihoods


class SortedSpikesDetector(_BaseDetector):
    def __init__(
        self,
        place_bin_size=2.0,
        position_range=None,
        track_graph=None,
        edge_order=None,
        edge_spacing=None,
        continuous_state_transition_type='random_walk',
        random_walk_variance=6.0,
        discrete_state_transition_type='infer_from_training_data',
        discrete_transition_diagonal=_DISCRETE_DIAGONAL,
        is_track_interior=None,
        infer_track_interior=True,
        spike_model_penalty=1E-5,
        spike_model_knot_spacing=12.5,
    ):
        super().__init__(
            place_bin_size,
            position_range,
            track_graph,
            edge_order,
            edge_spacing,
            continuous_state_transition_type,
            random_walk_variance,
            discrete_state_transition_type,
            discrete_transition_diagonal,
            is_track_interior,
            infer_track_interior,
        )
        self.spike_model_knot_spacing = spike_model_knot_spacing
        self.spike_model_penalty = spike_model_penalty

    def fit_place_fields(self, position, spikes, is_training=None):
        logger.info('Fitting place fields...')
        position = atleast_2d(np.asarray(position))
        spikes = np.asarray(spikes)
        if is_training is None:
            is_training = np.ones((spikes.shape[0],), dtype=bool)
        else:
            is_training = np.asarray(is_training)

        self.encoding_model_ = fit_spiking_likelihood(
            position,
            spikes,
            is_training,
            self.place_bin_centers_,
            self.place_bin_edges_,
            self.is_track_interior_.ravel(order='F'),
            self.spike_model_penalty,
            self.spike_model_knot_spacing)

    def plot_place_fields(
        self,
        sampling_frequency=1,
        col_wrap=5,
        axes=None
    ):
        """Plot the place fields from the fitted spiking data.

        Parameters
        ----------
        ax : matplotlib axes or None, optional
        sampling_frequency : float, optional

        """
        place_conditional_intensity = (
            self.encoding_model_
            .keywords['place_conditional_intensity']).squeeze()
        n_neurons = place_conditional_intensity.shape[1]
        n_rows = np.ceil(n_neurons / col_wrap).astype(np.int)

        if axes is None:
            fig, axes = plt.subplots(n_rows, col_wrap, sharex=True,
                                     figsize=(col_wrap * 2, n_rows * 2),
                                     constrained_layout=True)

        for ind, ax in enumerate(axes.flat):
            if ind < n_neurons:
                mask = np.ones_like(self.place_bin_centers_.squeeze())
                mask[~self.is_track_interior_.ravel(order='F')] = np.nan
                ax.plot(self.place_bin_centers_,
                        place_conditional_intensity[:, ind] *
                        sampling_frequency * mask, color='black', linewidth=1,
                        label='fitted model')
                ax.set_title(f'Neuron #{ind + 1}')
                ax.set_ylabel('Spikes / s')
                ax.set_xlabel('Position')
            else:
                ax.axis('off')

    @staticmethod
    def plot_spikes(position, spikes, is_training=None, sampling_frequency=1,
                    col_wrap=5, bins='auto'):
        if is_training is None:
            is_training = np.ones((spikes.shape[0], ), dtype=bool)
        else:
            is_training = np.asarray(is_training.copy()).squeeze()
        position = np.asarray(position.copy()).squeeze()[is_training]
        position = atleast_2d(np.asarray(position))
        spikes = np.asarray(spikes.copy())[is_training]

        position_occupancy, bin_edges = np.histogram(position, bins=bins)
        bin_size = np.diff(bin_edges)[0]

        time_ind, neuron_ind = np.nonzero(spikes)
        n_neurons = spikes.shape[1]

        n_rows = np.ceil(n_neurons / col_wrap).astype(np.int)

        fig, axes = plt.subplots(n_rows, col_wrap, sharex=True,
                                 figsize=(col_wrap * 2, n_rows * 2))

        for ind, ax in enumerate(axes.flat):
            if ind < n_neurons:
                hist, _ = np.histogram(position[time_ind[neuron_ind == ind]],
                                       bins=bin_edges)
                rate = sampling_frequency * hist / position_occupancy
                ax.bar(bin_edges[:-1], rate, width=bin_size)
                ax.set_title(f'Neuron #{ind + 1}')
                ax.set_ylabel('Spikes / s')
                ax.set_xlabel('Position')
            else:
                ax.axis('off')

        plt.tight_layout()

        return axes

    def fit(self, position, spikes, is_training=None, refit=False):
        position = atleast_2d(np.asarray(position))
        spikes = np.asarray(spikes)
        if is_training is None:
            is_training = np.ones((position.shape[0],), dtype=bool)
        else:
            is_training = np.asarray(is_training)

        if not refit:
            self.fit_place_grid(
                position=position,
                track_graph=self.track_graph,
                edge_order=self.edge_order,
                edge_spacing=self.edge_spacing,
                infer_track_interior=self.infer_track_interior,
            )

            self.fit_discrete_state_transition(
                discrete_transition_diagonal=self.discrete_transition_diagonal,
                is_training=is_training
            )
            self.fit_continuous_state_transition(
                position=position,
                is_training=is_training,
            )

        self.fit_place_fields(
            position,
            spikes,
            is_training
        )

        return self

    def predict(
        self,
        position,
        spikes,
        time=None,
        is_compute_acausal=True,
        set_no_spike_to_equally_likely=False,
        use_gpu=False,
        store_likelihood=False
    ):
        position = atleast_2d(np.asarray(position))
        spikes = np.asarray(spikes)

        logger.info('Estimating likelihood...')
        likelihood = self.encoding_model_(
            spikes=spikes,
            position=position,
            set_no_spike_to_equally_likely=set_no_spike_to_equally_likely)

        if store_likelihood:
            self.likelihood_ = likelihood

        return self._get_results(
            position,
            likelihood,
            time=time,
            is_compute_acausal=is_compute_acausal,
            use_gpu=use_gpu,
        )


class ClusterlessDetector(_BaseDetector):
    def __init__(
        self,
        place_bin_size=2.0,
        position_range=None,
        track_graph=None,
        edge_order=None,
        edge_spacing=None,
        continuous_state_transition_type='random_walk',
        random_walk_variance=6.0,
        discrete_state_transition_type='infer_from_training_data',
        discrete_transition_diagonal=_DISCRETE_DIAGONAL,
        is_track_interior=None,
        infer_track_interior=True,
        clusterless_algorithm='multiunit_likelihood',
        clusterless_algorithm_params=_DEFAULT_CLUSTERLESS_MODEL_KWARGS
    ):
        super().__init__(
            place_bin_size,
            position_range,
            track_graph,
            edge_order,
            edge_spacing,
            continuous_state_transition_type,
            random_walk_variance,
            discrete_state_transition_type,
            discrete_transition_diagonal,
            is_track_interior,
            infer_track_interior,
        )
        self.clusterless_algorithm = clusterless_algorithm
        self.clusterless_algorithm_params = clusterless_algorithm_params

    def fit_multiunits(
            self,
            position,
            multiunits,
            is_training=None,
    ):
        '''

        Parameters
        ----------
        position : array_like, shape (n_time, n_position_dims)
        multiunits : array_like, shape (n_time, n_features, n_electrodes)
        is_training : None or array_like, shape (n_time,)

        '''
        logger.info('Fitting multiunits...')
        position = atleast_2d(np.asarray(position))
        multiunits = np.asarray(multiunits)

        if is_training is None:
            is_training = np.ones((position.shape[0],), dtype=bool)
        else:
            is_training = np.asarray(is_training)

        if self.clusterless_algorithm == 'multiunit_likelihood':
            self.encoding_model_ = fit_multiunit_likelihood(
                position=position,
                multiunits=multiunits,
                is_training=is_training,
                place_bin_centers=self.place_bin_centers_,
                is_track_interior=self.is_track_interior_.ravel(order='F'),
                **self.clusterless_algorithm_params,
            )
        elif self.clusterless_algorithm == 'multiunit_likelihood_integer':
            self.encoding_model_ = fit_multiunit_likelihood_integer(
                position=position,
                multiunits=multiunits,
                is_training=is_training,
                place_bin_centers=self.place_bin_centers_,
                is_track_interior=self.is_track_interior_.ravel(order='F'),
                **self.clusterless_algorithm_params
            )
        elif self.clusterless_algorithm == 'multiunit_likelihood_integer_gpu':
            self.encoding_model_ = fit_multiunit_likelihood_gpu(
                position=position,
                multiunits=multiunits,
                is_training=is_training,
                place_bin_centers=self.place_bin_centers_,
                is_track_interior=self.is_track_interior_.ravel(order='F'),
                **self.clusterless_algorithm_params
            )
        else:
            raise NotImplementedError

    def fit(self, position, multiunits, is_training=None, refit=False):
        position = atleast_2d(np.asarray(position))
        multiunits = np.asarray(multiunits)
        if is_training is None:
            is_training = np.ones((position.shape[0],), dtype=bool)
        else:
            is_training = np.asarray(is_training)

        if not refit:
            self.fit_place_grid(
                position=position,
                track_graph=self.track_graph,
                edge_order=self.edge_order,
                edge_spacing=self.edge_spacing,
                infer_track_interior=self.infer_track_interior,
            )

            self.fit_discrete_state_transition(
                discrete_transition_diagonal=self.discrete_transition_diagonal,
                is_training=is_training
            )
            self.fit_continuous_state_transition(
                position=position,
                is_training=is_training,
            )

        self.fit_multiunits(
            position,
            multiunits,
            is_training
        )

        return self

    def predict(
        self,
        position,
        multiunits,
        time=None,
        is_compute_acausal=True,
        set_no_spike_to_equally_likely=False,
        use_gpu=False,
        store_likelihood=False
    ):
        position = atleast_2d(np.asarray(position))
        multiunits = np.asarray(multiunits)

        logger.info('Estimating likelihood...')
        likelihood = self.encoding_model_(
            multiunits=multiunits,
            position=position,
            set_no_spike_to_equally_likely=set_no_spike_to_equally_likely)

        if store_likelihood:
            self.likelihood_ = likelihood

        return self._get_results(
            position,
            likelihood,
            time=time,
            is_compute_acausal=is_compute_acausal,
            use_gpu=use_gpu,
        )



'''Tools for simulating place field spiking'''
import numpy as np
from replay_identification.bins import atleast_2d
from scipy.stats import multivariate_normal, norm


def simulate_time(n_samples, sampling_frequency):
    return np.arange(n_samples) / sampling_frequency


def simulate_linear_distance(time, track_height, running_speed=10):
    half_height = (track_height / 2)
    return (half_height * np.sin(2 * np.pi * time / running_speed - np.pi / 2)
            + half_height)


def simulate_linear_distance_with_pauses(time, track_height, running_speed=10,
                                         pause=0.5, sampling_frequency=1):
    linear_distance = simulate_linear_distance(
        time, track_height, running_speed)
    peaks = np.nonzero(linear_distance == track_height)[0]
    n_pause_samples = int(pause * sampling_frequency)
    pause_linear_distance = np.zeros(
        (time.size + n_pause_samples * peaks.size,))
    pause_ind = (peaks[:, np.newaxis] + np.arange(n_pause_samples))
    pause_ind += np.arange(peaks.size)[:, np.newaxis] * n_pause_samples

    pause_linear_distance[pause_ind.ravel()] = track_height
    pause_linear_distance[pause_linear_distance == 0] = linear_distance
    return pause_linear_distance[:time.size]


def get_trajectory_direction(linear_distance):
    is_inbound = np.insert(np.diff(linear_distance) < 0, 0, False)
    return np.where(is_inbound, 'Inbound', 'Outbound'), is_inbound


def simulate_poisson_spikes(rate, sampling_frequency):
    return 1.0 * (np.random.poisson(rate / sampling_frequency) > 0)


def create_place_field(
    place_field_mean, linear_distance, sampling_frequency, is_condition=None,
        place_field_std_deviation=12.5, max_firing_rate=20,
        baseline_firing_rate=0.001):
    if is_condition is None:
        is_condition = np.ones_like(linear_distance, dtype=bool)
    field_firing_rate = norm(
        place_field_mean, place_field_std_deviation).pdf(linear_distance)
    field_firing_rate /= np.nanmax(field_firing_rate)
    field_firing_rate[~is_condition] = 0
    return baseline_firing_rate + max_firing_rate * field_firing_rate


def simulate_place_field_firing_rate(means, position, max_rate=15,
                                     variance=10, is_condition=None):
    '''Simulates the firing rate of a neuron with a place field at `means`.

    Parameters
    ----------
    means : ndarray, shape (n_position_dims,)
    position : ndarray, shape (n_time, n_position_dims)
    max_rate : float, optional
    variance : float, optional
    is_condition : None or ndarray, (n_time,)

    Returns
    -------
    firing_rate : ndarray, shape (n_time,)

    '''
    if is_condition is None:
        is_condition = np.ones(position.shape[0], dtype=bool)
    position = atleast_2d(position)
    firing_rate = multivariate_normal(means, variance).pdf(position)
    firing_rate /= firing_rate.max()
    firing_rate *= max_rate
    firing_rate[~is_condition] = 0.0

    return firing_rate


def simulate_neuron_with_place_field(means, position, max_rate=15, variance=36,
                                     sampling_frequency=500,
                                     is_condition=None):
    '''Simulates the spiking of a neuron with a place field at `means`.

    Parameters
    ----------
    means : ndarray, shape (n_position_dims,)
    position : ndarray, shape (n_time, n_position_dims)
    max_rate : float, optional
    variance : float, optional
    sampling_frequency : float, optional
    is_condition : None or ndarray, (n_time,)

    Returns
    -------
    spikes : ndarray, shape (n_time,)

    '''
    firing_rate = simulate_place_field_firing_rate(
        means, position, max_rate, variance, is_condition)
    return simulate_poisson_spikes(firing_rate, sampling_frequency)



from functools import partial
from logging import getLogger

import numpy as np
import pandas as pd
from patsy import NAAction, build_design_matrices, dmatrices
from regularized_glm import penalized_IRLS
from scipy.special import logsumexp
from statsmodels.api import families
from statsmodels.tsa.tsatools import lagmat

FAMILY = families.Binomial()

logger = getLogger(__name__)


def fit_discrete_state_transition(speed, is_replay, penalty=1E-5,
                                  speed_knots=None, diagonal=None):
    """Estimate the predicted probablity of replay given speed and whether
    it was a replay in the previous time step.

    p(I_t | I_t-1, v_t-1)

    p_I_0, p_I_1 in Long Tao's code

    Parameters
    ----------
    speed : ndarray, shape (n_time,)
    is_replay : boolean ndarray, shape (n_time,)
    speed_knots : ndarray, shape (n_knots,)

    Returns
    -------
    probability_replay : ndarray, shape (n_time, 2)

    """
    data = pd.DataFrame({
        'is_replay': is_replay.astype(np.float64),
        'lagged_is_replay': lagmat(
            is_replay, maxlag=1).astype(np.float64).squeeze(),
        'lagged_speed': lagmat(speed, maxlag=1).squeeze()
    }).dropna()

    if speed_knots is None:
        speed_mid_point = np.nanmedian(speed[speed > 10])
        speed_knots = [1., 2., 3., speed_mid_point]

    MODEL_FORMULA = (
        'is_replay ~ 1 + lagged_is_replay + '
        'cr(lagged_speed, knots=speed_knots, constraints="center")')
    response, design_matrix = dmatrices(MODEL_FORMULA, data)
    penalty = np.ones((design_matrix.shape[1],)) * penalty
    penalty[0] = 0.0
    fit = penalized_IRLS(design_matrix, response, family=FAMILY,
                         penalty=penalty)
    if np.isnan(fit.AIC):
        logger.error("Discrete state transition failed to fit properly. "
                     "Try specifying `speed_knots`")
    return partial(predict_probability, design_matrix=design_matrix,
                   coefficients=fit.coefficients)


def fit_discrete_state_transition_no_speed(
        speed, is_replay, penalty=1E-5, speed_knots=None, diagonal=None):
    """Estimate the predicted probablity of replay and whether
    it was a replay in the previous time step.

    p(I_t | I_t-1, v_t-1)

    p_I_0, p_I_1 in Long Tao's code

    Parameters
    ----------
    speed : ndarray, shape (n_time,)
    is_replay : boolean ndarray, shape (n_time,)
    speed_knots : ndarray, shape (n_knots,)

    Returns
    -------
    probability_replay : ndarray, shape (n_time, 2)

    """
    data = pd.DataFrame({
        'is_replay': is_replay.astype(np.float64),
        'lagged_is_replay': lagmat(
            is_replay, maxlag=1).astype(np.float64).squeeze(),
        'lagged_speed': lagmat(speed, maxlag=1).squeeze()
    }).dropna()

    MODEL_FORMULA = 'is_replay ~ 1 + lagged_is_replay'
    response, design_matrix = dmatrices(MODEL_FORMULA, data)
    penalty = np.ones((design_matrix.shape[1],)) * penalty
    penalty[0] = 0.0
    fit = penalized_IRLS(design_matrix, response, family=FAMILY,
                         penalty=penalty)
    return partial(predict_probability, design_matrix=design_matrix,
                   coefficients=fit.coefficients)


def constant_transition(
        speed, is_replay, penalty=1E-5, speed_knots=None, diagonal=None):
    '''
    discrete_state_transition[0] = Pr(I_{k} = 1 | I_{k-1} = 0)
    discrete_state_transition[1] = Pr(I_{k} = 1 | I_{k-1} = 1)

    0 : from local to non_local
    1 : from non_local to non_local
    '''

    if diagonal is None:
        diagonal = np.array([0.00003, 0.98])
    return partial(_constant_probability, diagonal=diagonal)


def make_design_matrix(lagged_is_replay, lagged_speed, design_matrix):
    predict_data = {
        'lagged_is_replay': lagged_is_replay * np.ones_like(lagged_speed),
        'lagged_speed': lagged_speed
    }
    return build_design_matrices(
        [design_matrix.design_info], predict_data,
        NA_action=NAAction(NA_types=[]))[0]


def make_design_matrix_no_speed(lagged_is_replay, lagged_speed, design_matrix):
    predict_data = {
        'lagged_is_replay': lagged_is_replay * np.ones_like(lagged_speed),
    }
    return build_design_matrices(
        [design_matrix.design_info], predict_data,
        NA_action=NAAction(NA_types=[]))[0]


def predict_probability(lagged_speed, design_matrix, coefficients):
    """Predict probability of replay state given speed and whether it was a
    replay in the previous time step.

    Parameters
    ----------
    design_matrix : patsy design matrix
    fit : statsmodels fitted model
    speed : ndarray, shape (n_time,)

    Returns
    -------
    replay_probability : ndarray, shape (n_time, 2)

    """
    no_previous_replay_design_matrix = make_design_matrix(
        0, lagged_speed, design_matrix)

    previous_replay_design_matrix = make_design_matrix(
        1, lagged_speed, design_matrix)

    replay_probability = np.stack(
        (predict(no_previous_replay_design_matrix, coefficients),
         predict(previous_replay_design_matrix, coefficients)),
        axis=1)

    replay_probability[np.isnan(replay_probability)] = 0.0

    return replay_probability


def predict_probability_no_speed(lagged_speed, design_matrix, coefficients):
    """Predict probability of replay state given speed and whether it was a
    replay in the previous time step.

    Parameters
    ----------
    design_matrix : patsy design matrix
    fit : statsmodels fitted model
    speed : ndarray, shape (n_time,)

    Returns
    -------
    replay_probability : ndarray, shape (n_time, 2)

    """
    no_previous_replay_design_matrix = make_design_matrix_no_speed(
        0, lagged_speed, design_matrix)

    previous_replay_design_matrix = make_design_matrix_no_speed(
        1, lagged_speed, design_matrix)

    replay_probability = np.stack(
        (predict(no_previous_replay_design_matrix, coefficients),
         predict(previous_replay_design_matrix, coefficients)),
        axis=1)

    replay_probability[np.isnan(replay_probability)] = 0.0

    return replay_probability


def _constant_probability(lagged_speed, diagonal=None):
    if diagonal is None:
        diagonal = np.array([0.00003, 0.98])
    n_time = lagged_speed.shape[0]
    replay_probability = np.zeros((n_time, 2))
    replay_probability[:, 0] = diagonal[0]
    replay_probability[:, 1] = diagonal[1]

    return replay_probability


def predict(design_matrix, coefficients):
    return FAMILY.link.inverse(design_matrix @ np.squeeze(coefficients))


def estimate_discrete_state_transition(detector, results):
    EPS = 1e-32
    try:
        causal_prob = np.log(
            results.causal_posterior.sum('position').values + EPS)
        acausal_prob = np.log(
            results.acausal_posterior.sum('position').values + EPS)
    except ValueError:
        causal_prob = np.log(
            results.causal_posterior.sum(['x_position', 'y_position']).values +
            EPS)
        acausal_prob = np.log(
            results.acausal_posterior.sum(['x_position', 'y_position']).values
            + EPS)

    try:
        transition = detector.discrete_state_transition_(np.arange(1))[0]
        old_discrete_state_transition = np.log(np.asarray(
            [[1 - transition[0], transition[0]],
             [1 - transition[1], transition[1]]]))
    except TypeError:
        old_discrete_state_transition = np.log(
            detector.discrete_state_transition_)

    n_states = old_discrete_state_transition.shape[0]

    new_log_discrete_state_transition = np.empty((n_states, n_states))
    for i in range(n_states):
        for j in range(n_states):
            new_log_discrete_state_transition[i, j] = logsumexp(
                old_discrete_state_transition[i, j] +
                causal_prob[:-1, i] +
                acausal_prob[1:, j] -
                causal_prob[1:, j])
            new_log_discrete_state_transition[i, j] -= logsumexp(
                acausal_prob[:-1, i])
    new_log_discrete_state_transition -= logsumexp(
        new_log_discrete_state_transition, axis=-1, keepdims=True)

    return np.exp(new_log_discrete_state_transition)


def make_discrete_state_transition_from_diagonal(diagonal):
    """Makes discrete state transition matrix.

    Parameters
    ----------
    n_states : int

    Returns
    -------
    discrete_state_transition : np.ndarray, shape (n_states, n_states)

    """
    n_states = len(diagonal)
    discrete_state_transition = np.identity(n_states) * diagonal
    is_off_diag = ~np.identity(n_states, dtype=bool)
    discrete_state_transition[is_off_diag] = (
        (1 - diagonal) / (n_states - 1))

    return discrete_state_transition


def infer_discrete_state_transition_from_training_data(
    is_non_local,
    penalty=1e-5
):
    data = pd.DataFrame({
        'is_non_local': is_non_local.astype(np.float64),
        'lagged_is_non_local': lagmat(
            is_non_local, maxlag=1).astype(np.float64).squeeze(),
    }).dropna()

    MODEL_FORMULA = 'is_non_local ~ 1 + lagged_is_non_local'
    response, design_matrix = dmatrices(MODEL_FORMULA, data)
    penalty = np.ones((design_matrix.shape[1],)) * penalty
    penalty[0] = 0.0
    fit = penalized_IRLS(design_matrix, response, family=families.Binomial(),
                         penalty=penalty)

    predict_data = {
        'lagged_is_non_local': np.asarray([0, 1]),
    }
    predict_design_matrix = build_design_matrices(
        [design_matrix.design_info], predict_data,
        NA_action=NAAction(NA_types=[]))[0]

    non_local_probability = families.Binomial().link.inverse(
        predict_design_matrix @ np.squeeze(fit.coefficients))

    non_local_probability[np.isnan(non_local_probability)] = 0.0

    return np.asarray(
        [[1 - non_local_probability[0], non_local_probability[0]],
         [1 - non_local_probability[1], non_local_probability[1]]])


_DISCRETE_STATE_TRANSITIONS = {
    'ripples_with_speed_threshold': fit_discrete_state_transition,
    'ripples_no_speed_threshold': fit_discrete_state_transition_no_speed,
    'constant': constant_transition,
}



"""Fitting and predicting the likelihood of non_local events based on place
field spiking patterns.
"""

import logging
from functools import partial
from logging import getLogger

import dask
import numpy as np
import scipy.stats
import statsmodels.api as sm
from dask.distributed import Client, get_client
from patsy import build_design_matrices, dmatrix
from statsmodels.api import families
from tqdm.autonotebook import tqdm

from .bins import atleast_2d, get_n_bins
from .core import scale_likelihood

logger = getLogger(__name__)


@dask.delayed
def fit_glm(response, design_matrix, is_training=None,
            penalty=None, tolerance=1E-5):
    if penalty is not None:
        penalty = np.ones((design_matrix.shape[1],)) * penalty
        penalty[0] = 0.0  # don't penalize the intercept
    else:
        penalty = np.finfo(np.float).eps

    glm = sm.GLM(
        response.squeeze(),
        design_matrix,
        family=families.Poisson(),
        var_weights=is_training.squeeze())

    return glm.fit_regularized(
        alpha=penalty,
        L1_wt=0,
        cnvrg_tol=tolerance)


def poisson_log_likelihood(spikes, conditional_intensity):
    """Probability of parameters given spiking at a particular time.

    Parameters
    ----------
    spikes : ndarray, shape (n_time,)
        Indicator of spike or no spike at current time.
    conditional_intensity : ndarray, shape (n_place_bins,)
        Instantaneous probability of observing a spike

    Returns
    -------
    poisson_log_likelihood : array_like, shape (n_time, n_place_bins)

    """
    return scipy.stats.poisson.logpmf(
        spikes,
        conditional_intensity + np.spacing(1))


def spiking_likelihood(
        spikes, position, design_matrix, place_field_coefficients,
        place_conditional_intensity, is_track_interior,
        set_no_spike_to_equally_likely=False):
    """Computes the likelihood of non-local and local events.

    Parameters
    ----------
    spikes : ndarray, shape (n_time, n_neurons)
    position : ndarray, shape (n_time,)
    design_matrix : ndarray, shape (n_time, n_coefficients)
    place_field_coefficients : ndarray, shape (n_coefficients, n_neurons)
    place_conditional_intensity : ndarray, shape (n_place_bins, n_neurons)
    time_bin_size : float, optional

    Returns
    -------
    spiking_likelihood : ndarray, shape (n_time, n_place_bins,)

    """
    local_design_matrix = make_spline_predict_matrix(
        design_matrix.design_info, position)
    local_conditional_intensity = get_firing_rate(
        local_design_matrix, place_field_coefficients, sampling_frequency=1)
    n_time = spikes.shape[0]
    n_place_bins = place_conditional_intensity.shape[0]
    spiking_log_likelihood = np.zeros((n_time, 2, n_place_bins))

    # Non-Local
    spiking_log_likelihood[:, 1, :] = (combined_likelihood(
        spikes.T[..., np.newaxis],
        place_conditional_intensity.T[:, np.newaxis, :]))

    # Local
    spiking_log_likelihood[:, 0, :] = (combined_likelihood(
        spikes.T, local_conditional_intensity.T)
    )

    if set_no_spike_to_equally_likely:
        no_spike = np.isclose(spikes.sum(axis=1), 0.0)
        spiking_log_likelihood[no_spike] = 0.0

    is_track_interior = is_track_interior.ravel(order='F')
    spiking_log_likelihood[:, :, ~is_track_interior] = np.nan

    return scale_likelihood(spiking_log_likelihood)


def combined_likelihood(spikes, conditional_intensity):
    n_time = spikes.shape[1]
    n_bins = (conditional_intensity.shape[-1]
              if conditional_intensity.ndim > 2 else 1)
    log_likelihood = np.zeros((n_time, n_bins))

    for is_spike, ci in zip(tqdm(spikes, desc='neurons'),
                            conditional_intensity):
        log_likelihood += atleast_2d(
            poisson_log_likelihood(is_spike, ci))

    return log_likelihood


def make_spline_design_matrix(position, place_bin_edges, knot_spacing=10):
    inner_knots = []
    for pos, edges in zip(position.T, place_bin_edges.T):
        n_points = get_n_bins(edges, bin_size=knot_spacing)
        knots = np.linspace(edges.min(), edges.max(), n_points)[1:-1]
        knots = knots[(knots > pos.min()) & (knots < pos.max())]
        inner_knots.append(knots)

    inner_knots = np.meshgrid(*inner_knots)

    data = {}
    formula = '1 + te('
    for ind in range(position.shape[1]):
        formula += f'cr(x{ind}, knots=inner_knots[{ind}])'
        formula += ', '
        data[f'x{ind}'] = position[:, ind]

    formula += 'constraints="center")'

    return dmatrix(formula, data)


def make_spline_predict_matrix(design_info, position):
    position = atleast_2d(position)
    is_nan = np.any(np.isnan(position), axis=1)
    position[is_nan] = 0.0

    predict_data = {}
    for ind in range(position.shape[1]):
        predict_data[f'x{ind}'] = position[:, ind]

    design_matrix = build_design_matrices(
        [design_info], predict_data)[0]
    design_matrix[is_nan] = np.nan

    return design_matrix


def get_firing_rate(design_matrix, coefficients, sampling_frequency=1):
    rate = np.exp(design_matrix @ coefficients) * sampling_frequency
    rate[np.isnan(rate)] = np.spacing(1)
    return rate


def fit_spiking_likelihood(position, spikes, is_training,
                           place_bin_centers, place_bin_edges,
                           is_track_interior, penalty=1E1,
                           knot_spacing=30):
    """Estimate the place field model.

    Parameters
    ----------
    position : ndarray, shape (n_time,)
    spikes : ndarray, shape (n_time, n_neurons)
    place_bin_centers : ndarray, shape (n_place_bins,)
    penalty : float, optional
    time_bin_size : float, optional

    Returns
    -------
    spiking_likelihood : function

    """
    if np.any(np.ptp(place_bin_edges, axis=0) <= knot_spacing):
        logging.warning("Range of position is smaller than knot spacing.")

    is_training = np.asarray(is_training).astype(float)
    include = ~np.isclose(is_training, 0.0) & ~np.any(
        np.isnan(position), axis=1)
    is_training = is_training[include]
    position = position[include]
    spikes = spikes[include]

    design_matrix = make_spline_design_matrix(
        position, place_bin_edges, knot_spacing)
    try:
        client = get_client()
    except ValueError:
        client = Client()
    dm = client.scatter(np.asarray(design_matrix), broadcast=True)

    place_field_coefficients = [
        fit_glm(is_spike,
                dm,
                is_training,
                penalty).params
        for is_spike in spikes.T]
    place_field_coefficients = np.stack(
        dask.compute(*place_field_coefficients), axis=1)

    predict_matrix = make_spline_predict_matrix(
        design_matrix.design_info, place_bin_centers)
    place_conditional_intensity = get_firing_rate(
        predict_matrix, place_field_coefficients, sampling_frequency=1)

    return partial(
        spiking_likelihood,
        design_matrix=design_matrix,
        place_field_coefficients=place_field_coefficients,
        place_conditional_intensity=place_conditional_intensity,
        is_track_interior=is_track_interior)



import dask
import networkx as nx
import numba
import numpy as np
from scipy.stats import multivariate_normal

SQRT_2PI = np.float64(np.sqrt(2.0 * np.pi))


def _setup_distance(linear_position, nodes_df):
    linear_position = linear_position.squeeze()
    bin_ind = np.searchsorted(nodes_df.linear_position.values, linear_position)
    is_same_edge = (nodes_df.iloc[bin_ind - 1].edge_id.values ==
                    nodes_df.iloc[bin_ind].edge_id.values)

    left_node_ind = bin_ind - 1
    right_node_ind = bin_ind

    right_node_ind[~is_same_edge] = left_node_ind[~is_same_edge] = np.argmin(
        np.abs(linear_position[~is_same_edge, np.newaxis] -
               nodes_df.linear_position.values), axis=1)

    left_node_id = nodes_df.node_ids.values[left_node_ind]
    right_node_id = nodes_df.node_ids.values[right_node_ind]
    left_dist = np.abs(
        nodes_df.linear_position.values[left_node_ind] - linear_position)
    right_dist = np.abs(
        nodes_df.linear_position.values[left_node_ind] - linear_position)

    return left_node_id, right_node_id, left_dist, right_dist


def get_distance2(track_graph, left_node_id, right_node_id, left_dist,
                  right_dist, place_bin_center_to_node_id):
    nx.add_path(track_graph, [left_node_id, 'a'], distance=left_dist)
    nx.add_path(track_graph, ['a', right_node_id], distance=right_dist)
    dist = nx.single_source_dijkstra_path_length(
        track_graph, source='a', weight="distance")
    track_graph.remove_node('a')
    r = np.array([dist[id] for id in place_bin_center_to_node_id])

    return r


@dask.delayed
def batch_distance(track_graph, left_node_id, right_node_id,
                   left_dist, right_dist, place_bin_center_to_node_id):
    copy_graph = track_graph.copy()

    return np.stack(
        [get_distance2(copy_graph, l_id, r_id, l_d, r_d,
                       place_bin_center_to_node_id)
         for l_id, r_id, l_d, r_d in zip(left_node_id, right_node_id,
                                         left_dist, right_dist)], axis=0)


def batch(n_samples, batch_size=1):
    for ind in range(0, n_samples, batch_size):
        yield range(ind, min(ind + batch_size, n_samples))


def convert_linear_position_to_track_distances(
        linear_position, track_graph, nodes_df, place_bin_center_to_node_id):
    '''

    Parameters
    ----------

    Returns
    -------
    track_distances : np.ndarray, shape (n_time, n_place_bins)

    '''
    (left_node_id,
     right_node_id,
     left_dist,
     right_dist) = _setup_distance(linear_position, nodes_df)

    n_time = linear_position.shape[0]
    track_distances = []
    for time_ind in batch(n_time, batch_size=10_000):
        track_distances.append(
            batch_distance(
                track_graph, left_node_id[time_ind], right_node_id[time_ind],
                left_dist[time_ind], right_dist[time_ind],
                place_bin_center_to_node_id))

    return np.concatenate(
        dask.compute(*track_distances, scheduler='processes'), axis=0)


def get_gaussian_track_distances(track_distances, variance=8):
    return multivariate_normal(mean=0, cov=variance).pdf(
        track_distances.ravel()).reshape(track_distances.shape)


@numba.njit(nogil=True, cache=False, parallel=True, error_model='numpy')
def numba_product(eval_point, samples, bandwidths):
    '''
    Parameters
    ----------
    eval_point : np.ndarray, shape (n_marks,)
    samples : np.ndarray, shape (n_train, n_marks)
    bandwidths : np.ndarray, shape (n_marks,)
    Returns
    -------
    product_kernel : shape (n_train,)

    '''
    n_samples, n_bandwidths = samples.shape
    product_kernel = np.ones((n_samples,))

    for j in numba.prange(n_samples):
        for k in range(n_bandwidths):
            bandwidth = bandwidths[k]
            sample = samples[j, k]
            product_kernel[j] *= (np.exp(
                -0.5 * ((eval_point[k] - sample) / bandwidth)**2) /
                (bandwidth * SQRT_2PI)) / bandwidth

    return product_kernel


def get_kde(test_multiunit, train_multiunit, is_track_interior, bandwidths,
            gaussian_track_distances):
    '''

    Parameters
    ----------
    test_multiunit : np.ndarray, shape (n_test, n_marks)
    train_multiunit : np.ndarray, shape (n_train, n_marks)
    is_track_interior : np.ndarray, shape (n_bins, 1)
    bandwidths : np.ndarray, shape (n_marks,)
    gaussian_track_distances : np.ndarray, shape (n_train, n_bins)

    Returns
    -------
    kde : np.ndarray, shape (n_test, n_bins)

    '''
    n_test, n_bins = test_multiunit.shape[0], gaussian_track_distances.shape[1]
    kde = np.zeros((n_test, n_bins))
    for ind in range(n_test):
        kde[ind, is_track_interior] = (
            numba_product(test_multiunit[ind], train_multiunit, bandwidths) @
            gaussian_track_distances)

    return kde


@numba.njit(nogil=True, cache=False, parallel=True)
def numba_kde(eval_points, samples, bandwidths, precalculated_kernel):
    n_eval_points, n_bandwidths = eval_points.shape
    n_grid = precalculated_kernel.shape[1]
    result = np.zeros((n_eval_points, n_grid))
    n_samples = len(samples)

    for eval_ind in range(n_eval_points):
        for sample_ind in range(n_samples):
            product_kernel = 1.0
            for bandwidth_ind in range(n_bandwidths):
                bandwidth = bandwidths[bandwidth_ind]
                eval_point = eval_points[eval_ind, bandwidth_ind]
                sample = samples[sample_ind, bandwidth_ind]
                product_kernel *= (np.exp(
                    -0.5 * ((eval_point - sample) / bandwidth)**2) /
                    (bandwidth * SQRT_2PI)) / bandwidth
            for grid_ind in range(n_grid):
                result[eval_ind, grid_ind] += (
                    product_kernel * precalculated_kernel[sample_ind, grid_ind]
                    / n_samples)

    return result


@numba.njit(nogil=True, cache=False, parallel=True, fastmath=True)
def gaussian_kernel(eval_point, bandwidths, sample, n_bandwidths):
    product_kernel = 1.0
    for k in range(n_bandwidths):
        bandwidth = bandwidths[k]
        product_kernel *= (np.exp(
            -0.5 * ((eval_point[k] - sample[k]) / bandwidth)**2) /
            (bandwidth * SQRT_2PI)) / bandwidth
    return product_kernel



try:
    import cupy as cp
except ImportError:
    import numpy as cp
import numpy as np
from numba import njit


def replace_NaN(x):
    x[np.isnan(x)] = 1
    return x


def return_None(*args, **kwargs):
    return None


@njit(cache=True, nogil=True)
def normalize_to_probability(distribution):
    '''Ensure the distribution integrates to 1 so that it is a probability
    distribution
    '''
    return distribution / np.nansum(distribution)


@njit(cache=True, nogil=True, error_model='numpy')
def _causal_classifier(likelihood, movement_state_transition, discrete_state_transition,
                       observed_position_bin, uniform):
    '''
    Parameters
    ----------
    likelihood : ndarray, shape (n_time, ...)
    movement_state_transition : ndarray, shape (n_position_bins,
                                                n_position_bins)
    discrete_state_transition : ndarray, shape (n_time, 2)
        discrete_state_transition[k, 0] = Pr(I_{k} = 1 | I_{k-1} = 0, v_{k})
        discrete_state_transition[k, 1] = Pr(I_{k} = 1 | I_{k-1} = 1, v_{k})
    observed_position_bin : ndarray, shape (n_time,)
        Which position bin is the animal in.
    position_bin_size : float

    Returns
    -------
    posterior : ndarray, shape (n_time, 2, n_position_bins)
    state_probability : ndarray, shape (n_time, 2)
        state_probability[:, 0] = Pr(I_{1:T} = 0), Local
        state_probability[:, 1] = Pr(I_{1:T} = 1), Non-Local
    prior : ndarray, shape (n_time, 2, n_position_bins)

    '''
    n_position_bins = movement_state_transition.shape[0]
    n_time = likelihood.shape[0]
    n_states = 2

    posterior = np.zeros((n_time, n_states, n_position_bins))
    state_probability = np.zeros((n_time, n_states))

    # Initial Conditions
    posterior[0, 0, observed_position_bin[0]] = likelihood[0, 0, 0]
    norm = np.nansum(posterior[0])
    data_log_likelihood = np.log(norm)
    posterior[0] /= norm
    state_probability[0] = np.sum(posterior[0], axis=1)

    for k in np.arange(1, n_time):
        prior = np.zeros((n_states, n_position_bins))
        position_ind = observed_position_bin[k]
        # I_{k - 1} = 0, I_{k} = 0
        prior[0, position_ind] = (
            (1 - discrete_state_transition[k, 0]) * state_probability[k - 1, 0])
        # I_{k - 1} = 1, I_{k} = 0
        prior[0, position_ind] += (
            (1 - discrete_state_transition[k, 1]) * state_probability[k - 1, 1])

        # I_{k - 1} = 0, I_{k} = 1
        prior[1] = (discrete_state_transition[k, 0] * uniform *
                    state_probability[k - 1, 0])
        # I_{k - 1} = 1, I_{k} = 1
        prior[1] += (
            discrete_state_transition[k, 1] *
            (movement_state_transition.T @ posterior[k - 1, 1]))

        posterior[k] = prior * likelihood[k]
        norm = np.nansum(posterior[k])
        data_log_likelihood += np.log(norm)
        posterior[k] /= norm

        state_probability[k] = np.sum(posterior[k], axis=1)

    return posterior, state_probability, data_log_likelihood


@njit(cache=True, nogil=True, error_model='numpy')
def _acausal_classifier(filter_posterior, movement_state_transition,
                        discrete_state_transition, observed_position_bin, uniform):
    '''
    Parameters
    ----------
    filter_posterior : ndarray, shape (n_time, 2, n_position_bins)
    movement_state_transition : ndarray, shape (n_position_bins,
                                                n_position_bins)
    discrete_state_transition : ndarray, shape (n_time, 2)
        discrete_state_transition[k, 0] = Pr(I_{k} = 1 | I_{k-1} = 0, v_{k})
        discrete_state_transition[k, 1] = Pr(I_{k} = 1 | I_{k-1} = 1, v_{k})
    observed_position_bin : ndarray, shape (n_time,)
        Which position bin is the animal in.
    position_bin_size : float

    Returns
    -------
    smoother_posterior : ndarray, shape (n_time, 2, n_position_bins)
        p(x_{k + 1}, I_{k + 1} \vert H_{1:T})
    smoother_probability : ndarray, shape (n_time, 2)
        smoother_probability[:, 0] = Pr(I_{1:T} = 0)
        smoother_probability[:, 1] = Pr(I_{1:T} = 1)
    smoother_prior : ndarray, shape (n_time, 2, n_position_bins)
        p(x_{k + 1}, I_{k + 1} \vert H_{1:k})
    weights : ndarray, shape (n_time, 2, n_position_bins)
        \sum_{I_{k+1}} \int \Big[ \frac{p(x_{k+1} \mid x_{k}, I_{k}, I_{k+1}) *
        Pr(I_{k + 1} \mid I_{k}, v_{k}) * p(x_{k+1}, I_{k+1} \mid H_{1:T})}
        {p(x_{k + 1}, I_{k + 1} \mid H_{1:k})} \Big] dx_{k+1}
    '''  # noqa
    filter_probability = np.sum(filter_posterior, axis=2)

    smoother_posterior = np.zeros_like(filter_posterior)
    n_time, _, n_position_bins = filter_posterior.shape

    smoother_posterior[-1] = filter_posterior[-1].copy()

    for k in np.arange(n_time - 2, -1, -1):
        smoother_prior = np.zeros((2, n_position_bins))
        weights = np.zeros((2, n_position_bins))
        position_ind = observed_position_bin[k + 1]

        # Predict p(x_{k + 1}, I_{k + 1} \vert H_{1:k})
        # I_{k} = 0, I_{k + 1} = 0
        smoother_prior[0, position_ind] = (
            (1 - discrete_state_transition[k + 1, 0]) * filter_probability[k, 0])

        # I_{k} = 1, I_{k + 1} = 0
        smoother_prior[0, position_ind] += (
            (1 - discrete_state_transition[k + 1, 1]) * filter_probability[k, 1])

        # I_{k} = 0, I_{k + 1} = 1
        smoother_prior[1] = (
            discrete_state_transition[k + 1, 0] * uniform *
            filter_probability[k, 0])

        # I_{k} = 1, I_{k + 1} = 1
        smoother_prior[1] += (
            discrete_state_transition[k + 1, 1] *
            (movement_state_transition.T @ filter_posterior[k, 1]))

        # Update p(x_{k}, I_{k} \vert H_{1:k})
        ratio = np.exp(
            np.log(smoother_posterior[k + 1]) -
            np.log(smoother_prior + np.spacing(1)))
        integrated_ratio = np.sum(ratio, axis=1)
        # I_{k} = 0, I_{k + 1} = 0
        weights[0] = (
            (1 - discrete_state_transition[k + 1, 0]) * ratio[0, position_ind])

        # I_{k} = 0, I_{k + 1} = 1
        weights[0] += (
            uniform * discrete_state_transition[k + 1, 0] * integrated_ratio[1])

        # I_{k} = 1, I_{k + 1} = 0
        weights[1] = (
            (1 - discrete_state_transition[k + 1, 1]) * ratio[0, position_ind])

        # I_{k} = 1, I_{k + 1} = 1
        weights[1] += (
            discrete_state_transition[k + 1, 1] *
            ratio[1] @ movement_state_transition)

        smoother_posterior[k] = normalize_to_probability(
            weights * filter_posterior[k])

    smoother_probability = np.sum(smoother_posterior, axis=2)

    return smoother_posterior, smoother_probability


def scale_likelihood(log_likelihood):
    '''Scales the likelihood to its max value to prevent overflow and underflow.

    Parameters
    ----------
    log_likelihood : ndarray, shape (n_time, n_states, n_position_bins)

    Returns
    -------
    scaled_likelihood : ndarray, shape (n_time, n_states, n_position_bins)

    '''
    return np.exp(log_likelihood -
                  np.nanmax(log_likelihood, axis=(1, 2), keepdims=True))


def check_converged(loglik, previous_loglik, tolerance=1e-4):
    delta_loglik = abs(loglik - previous_loglik)
    avg_loglik = (abs(loglik) + abs(previous_loglik) + np.spacing(1)) / 2

    is_increasing = loglik - previous_loglik >= -1e-3
    is_converged = (delta_loglik / avg_loglik) < tolerance

    return is_converged, is_increasing


def _causal_classifier_gpu(likelihood, movement_state_transition, discrete_state_transition,
                           observed_position_bin, uniform):
    '''
    Parameters
    ----------
    likelihood : ndarray, shape (n_time, ...)
    movement_state_transition : ndarray, shape (n_position_bins,
                                                n_position_bins)
    discrete_state_transition : ndarray, shape (n_time, 2)
        discrete_state_transition[k, 0] = Pr(I_{k} = 1 | I_{k-1} = 0, v_{k})
        discrete_state_transition[k, 1] = Pr(I_{k} = 1 | I_{k-1} = 1, v_{k})
    observed_position_bin : ndarray, shape (n_time,)
        Which position bin is the animal in.
    position_bin_size : float

    Returns
    -------
    posterior : ndarray, shape (n_time, 2, n_position_bins)
    state_probability : ndarray, shape (n_time, 2)
        state_probability[:, 0] = Pr(I_{1:T} = 0)
        state_probability[:, 1] = Pr(I_{1:T} = 1)
    prior : ndarray, shape (n_time, 2, n_position_bins)

    '''

    likelihood = cp.asarray(likelihood, dtype=cp.float32)
    movement_state_transition = cp.asarray(
        movement_state_transition, dtype=cp.float32)
    discrete_state_transition = cp.asarray(
        discrete_state_transition, dtype=cp.float32)
    observed_position_bin = cp.asarray(observed_position_bin)
    uniform = cp.asarray(uniform, dtype=cp.float32)

    n_position_bins = movement_state_transition.shape[0]
    n_time = likelihood.shape[0]
    n_states = 2

    posterior = cp.zeros((n_time, n_states, n_position_bins), dtype=cp.float32)
    state_probability = cp.zeros((n_time, n_states), dtype=cp.float32)

    # Initial Conditions
    posterior[0, 0, observed_position_bin[0]] = likelihood[0, 0, 0]
    norm = cp.nansum(posterior[0])
    data_log_likelihood = cp.log(norm)
    posterior[0] /= norm
    state_probability[0] = cp.sum(posterior[0], axis=1)

    for k in np.arange(1, n_time):
        prior = cp.zeros((n_states, n_position_bins), dtype=cp.float32)
        position_ind = observed_position_bin[k]
        # I_{k - 1} = 0, I_{k} = 0
        prior[0, position_ind] = (
            (1 - discrete_state_transition[k, 0]) * state_probability[k - 1, 0])
        # I_{k - 1} = 1, I_{k} = 0
        prior[0, position_ind] += (
            (1 - discrete_state_transition[k, 1]) * state_probability[k - 1, 1])

        # I_{k - 1} = 0, I_{k} = 1
        prior[1] = (discrete_state_transition[k, 0] * uniform *
                    state_probability[k - 1, 0])
        # I_{k - 1} = 1, I_{k} = 1
        prior[1] += (
            discrete_state_transition[k, 1] *
            (movement_state_transition.T @ posterior[k - 1, 1]))

        posterior[k] = prior * likelihood[k]
        norm = cp.nansum(posterior[k])
        data_log_likelihood += cp.log(norm)
        posterior[k] /= norm

        state_probability[k] = cp.sum(posterior[k], axis=1)

    return (cp.asnumpy(posterior),
            cp.asnumpy(state_probability),
            data_log_likelihood)


def _acausal_classifier_gpu(filter_posterior, movement_state_transition,
                            discrete_state_transition, observed_position_bin, uniform):
    '''
    Parameters
    ----------
    filter_posterior : ndarray, shape (n_time, 2, n_position_bins)
    movement_state_transition : ndarray, shape (n_position_bins,
                                                n_position_bins)
    discrete_state_transition : ndarray, shape (n_time, 2)
        discrete_state_transition[k, 0] = Pr(I_{k} = 1 | I_{k-1} = 0, v_{k})
        discrete_state_transition[k, 1] = Pr(I_{k} = 1 | I_{k-1} = 1, v_{k})
    observed_position_bin : ndarray, shape (n_time,)
        Which position bin is the animal in.
    position_bin_size : float

    Returns
    -------
    smoother_posterior : ndarray, shape (n_time, 2, n_position_bins)
        p(x_{k + 1}, I_{k + 1} \vert H_{1:T})
    smoother_probability : ndarray, shape (n_time, 2)
        smoother_probability[:, 0] = Pr(I_{1:T} = 0)
        smoother_probability[:, 1] = Pr(I_{1:T} = 1)
    smoother_prior : ndarray, shape (n_time, 2, n_position_bins)
        p(x_{k + 1}, I_{k + 1} \vert H_{1:k})
    weights : ndarray, shape (n_time, 2, n_position_bins)
        \sum_{I_{k+1}} \int \Big[ \frac{p(x_{k+1} \mid x_{k}, I_{k}, I_{k+1}) *
        Pr(I_{k + 1} \mid I_{k}, v_{k}) * p(x_{k+1}, I_{k+1} \mid H_{1:T})}
        {p(x_{k + 1}, I_{k + 1} \mid H_{1:k})} \Big] dx_{k+1}
    '''  # noqa

    filter_posterior = cp.asarray(filter_posterior, dtype=cp.float32)
    movement_state_transition = cp.asarray(
        movement_state_transition, dtype=cp.float32)
    discrete_state_transition = cp.asarray(
        discrete_state_transition, dtype=cp.float32)
    observed_position_bin = cp.asarray(observed_position_bin)
    uniform = cp.asarray(uniform, dtype=cp.float32)
    EPS = cp.asarray(np.spacing(1), dtype=cp.float32)

    filter_probability = cp.sum(filter_posterior, axis=2)

    smoother_posterior = cp.zeros_like(filter_posterior)
    n_time, _, n_position_bins = filter_posterior.shape

    smoother_posterior[-1] = filter_posterior[-1].copy()

    for k in cp.arange(n_time - 2, -1, -1):
        smoother_prior = cp.zeros((2, n_position_bins), dtype=cp.float32)
        weights = cp.zeros((2, n_position_bins), dtype=cp.float32)

        position_ind = observed_position_bin[k + 1]

        # Predict p(x_{k + 1}, I_{k + 1} \vert H_{1:k})
        # I_{k} = 0, I_{k + 1} = 0
        smoother_prior[0, position_ind] = (
            (1 - discrete_state_transition[k + 1, 0]) * filter_probability[k, 0])

        # I_{k} = 1, I_{k + 1} = 0
        smoother_prior[0, position_ind] += (
            (1 - discrete_state_transition[k + 1, 1]) * filter_probability[k, 1])

        # I_{k} = 0, I_{k + 1} = 1
        smoother_prior[1] = (
            discrete_state_transition[k + 1, 0] * uniform *
            filter_probability[k, 0])

        # I_{k} = 1, I_{k + 1} = 1
        smoother_prior[1] += (
            discrete_state_transition[k + 1, 1] *
            (movement_state_transition.T @ filter_posterior[k, 1]))

        # Update p(x_{k}, I_{k} \vert H_{1:k})
        ratio = cp.exp(
            cp.log(smoother_posterior[k + 1]) -
            cp.log(smoother_prior + EPS))
        integrated_ratio = cp.sum(ratio, axis=1)
        # I_{k} = 0, I_{k + 1} = 0
        weights[0] = (
            (1 - discrete_state_transition[k + 1, 0]) * ratio[0, position_ind])

        # I_{k} = 0, I_{k + 1} = 1
        weights[0] += (
            uniform * discrete_state_transition[k + 1, 0] * integrated_ratio[1])

        # I_{k} = 1, I_{k + 1} = 0
        weights[1] = (
            (1 - discrete_state_transition[k + 1, 1]) * ratio[0, position_ind])

        # I_{k} = 1, I_{k + 1} = 1
        weights[1] += (
            discrete_state_transition[k + 1, 1] *
            ratio[1] @ movement_state_transition)

        smoother_posterior[k] = weights * filter_posterior[k]
        smoother_posterior[k] /= cp.nansum(smoother_posterior[k])

    smoother_probability = cp.sum(smoother_posterior, axis=2)

    return (cp.asnumpy(smoother_posterior),
            cp.asnumpy(smoother_probability))



import numpy as np
from numpy.random import randn
from replay_identification.simulate_LFP import simulate_LFP
from replay_identification.simulate_multiunit import simulate_multiunit
from replay_identification.simulate_spiking import (
    create_place_field, simulate_linear_distance_with_pauses,
    simulate_poisson_spikes, simulate_time)
from spectral_connectivity import Connectivity, Multitaper


def estimate_ripple_band_power(lfps, sampling_frequency):
    """Estimates the 200 Hz power of each LFP.

    Parameters
    ----------
    lfps : ndarray, shape (n_time, n_signals)
    sampling_frequency : float

    Returns
    -------
    ripple_band_power : ndarray (n_time, n_signals)

    """
    n_time = lfps.shape[0]
    m = Multitaper(lfps, sampling_frequency=sampling_frequency,
                   time_halfbandwidth_product=1,
                   time_window_duration=0.020,
                   time_window_step=1 / sampling_frequency)
    c = Connectivity.from_multitaper(m)
    closest_200Hz_freq_ind = np.argmin(np.abs(c.frequencies - 200))
    power = c.power()[..., closest_200Hz_freq_ind, :].squeeze()
    n_power_time = power.shape[0]
    unobserved = np.full((n_time - n_power_time, *power.shape[1:]), np.nan)
    return np.concatenate((power, unobserved))


def make_simulated_data():
    track_height = 170
    sampling_frequency = 1500
    n_samples = sampling_frequency * 65

    time = simulate_time(n_samples, sampling_frequency)
    linear_distance = (simulate_linear_distance_with_pauses(
        time, track_height, sampling_frequency=sampling_frequency, pause=3)
        + randn(*time.shape) * 1E-4)

    speed = np.abs(np.diff(linear_distance) / np.diff(time))
    speed = np.insert(speed, 0, 0.0)

    pause_ind = np.nonzero(
        np.diff(np.isclose(linear_distance, track_height, atol=1E-5)))[0] + 1

    pause_times = time[np.reshape(pause_ind, (-1, 2))]
    pause_width = np.diff(pause_times)[0]

    ripple_duration = 0.100

    lfps = np.stack((
        simulate_LFP(
            time, pause_times[:3, 0] + pause_width / 2, noise_amplitude=1.2,
            ripple_amplitude=1, ripple_width=ripple_duration),
        simulate_LFP(
            time, pause_times[:2, 0] + pause_width / 2, noise_amplitude=1.2,
            ripple_amplitude=1.1, ripple_width=ripple_duration)), axis=1
    )

    power = estimate_ripple_band_power(lfps, sampling_frequency)

    mid_ripple_time = pause_times[:3, 0] + pause_width / 2
    ripple_times = (mid_ripple_time +
                    np.array([-0.5, 0.5])[:, np.newaxis] * ripple_duration).T

    place_field_means = np.arange(0, 200, 25)

    place_fields = np.stack([create_place_field(
        place_field_mean, linear_distance, sampling_frequency)
        for place_field_mean in place_field_means])

    spikes = simulate_poisson_spikes(place_fields, sampling_frequency).T

    n_neurons = spikes.shape[1]
    n_samples_between_spikes = 20
    is_toward = np.array([0, 1], dtype=bool)

    for (start_time, end_time), flip in zip(ripple_times[[0, -1]], is_toward):
        is_ripple_time = (time >= start_time) & (time <= end_time)
        ripple_ind = np.nonzero(is_ripple_time)[0]
        spikes[is_ripple_time] = 0
        neuron_order = np.flip(np.arange(n_neurons),
                               axis=0) if flip else np.arange(n_neurons)
        replay_ind = (
            ripple_ind[0] +
            np.arange(0, n_neurons * n_samples_between_spikes,
                      n_samples_between_spikes),
            neuron_order)
        spikes[replay_ind] = 1

    mark_means = np.array([200, 125, 325, 275])
    place_field_means = np.stack((np.arange(0, 200, 50),
                                  np.arange(25, 200, 50)))

    multiunit = np.stack(
        [simulate_multiunit(
            place_field_means[0], mark_means, linear_distance,
            sampling_frequency),
         simulate_multiunit(
             place_field_means[1], mark_means, linear_distance,
             sampling_frequency),
         ], axis=-1)

    n_units = multiunit.shape[1]

    n_samples_between_spikes = 40
    is_toward = np.array([0, 1], dtype=bool)

    tetrode_ind = np.meshgrid(np.arange(multiunit.shape[1]), np.arange(
        multiunit.shape[2]))[1].flatten(order='F')
    mean_order = np.stack((mark_means, mark_means)).flatten(order='F')

    for start_time, end_time in ripple_times[[1]]:
        is_ripple_time = (time >= start_time) & (time <= end_time)
        clear_ind = (time >= start_time - 0.200) & (time <= end_time + 0.200)
        ripple_ind = np.nonzero(is_ripple_time)[0]
        multiunit[clear_ind] = np.nan
        time_ind = (ripple_ind[0] + np.arange(
            0, 2 * n_units * n_samples_between_spikes,
            n_samples_between_spikes))
        multiunit[time_ind, :, tetrode_ind] = mean_order[:, np.newaxis]

    is_replay = np.any(power > 0.0004, axis=1)

    return (is_replay, speed, linear_distance, power, spikes, time, multiunit)



# flake8: noqa
from replay_identification.decoders import ReplayDetector
from replay_identification.detectors import (ClusterlessDetector,
                                             SortedSpikesDetector)
from track_linearization import (get_linearized_position,
                                 make_actual_vs_linearized_position_movie,
                                 make_track_graph, plot_graph_as_1D,
                                 plot_track_graph)
